package feynmanplate;

import org.opensourcephysics.numerics.*;
import org.opensourcephysics.displayejs.*;
import java.awt.Color;
import java.awt.Font;

/**
 * <p>Title: RigidBody</p>
 * <p>Description: Abstract class. Its subclasses that define the transformBodyOnly()
 * method can draw themseves and compute their motion.</p>
 * <p>Copyright: Copyright (c) 2004</p>
 * @author Alexander Tomori, Boris Gazovic, Slavomir Tuleja
 * @version 0.01
 */
public abstract class RigidBody
    extends GroupDrawable
    implements ODE {

  double I, I3;
  double[] state = new double[7]; //{theta, phi, psi, thetaDot, phiDot, psiDot, t}
  ODEAdaptiveSolver solver;
  InteractiveArrow arrowX1, arrowX2, arrowX3; //unit vectors of a body-fixed frame
  InteractiveArrow arrowX, arrowY, arrowZ; //unit vectors of a laboratory frame
  InteractiveTrace traceX3, traceX1, traceX2; //traces traced out by end points of unit vectors x1, x2, x3
  InteractiveTrace traceX, traceY, traceZ; //traces traced out by end points of unit vectors x, y, z
  InteractiveArrow arrowOmega, arrowL; //vectors omega and L
  InteractiveTrace traceOmega, traceL; //traces traced out by end points of vectors omega nad L
  InteractiveText txtL, txtOmega, txtX1, txtX2, txtX3, txtX, txtY, txtZ;

  double[][] rm = new double[3][3]; //rotation matrix for transformation from body frame to laboratory frame
  double[][] rmT = new double[3][3]; //rotation matrix for transformation from laboratory frame to body frame

  boolean showRigidBody = true, showBodyFixedAxes = false, showLaboratoryAxes = false,
      showOmegaL = false, showTraces = false, isBodyFixedFrame = false;

  int numberOfCycles=0, numberOfCyclesOld=0;
  double startTracePsi;

  /**
   * Constructs RigidBody object.
   * @param I - principal moment of inertia in radial direction
   * @param I3 -principal moment of inertia in vertical symmetry axis direction
   */
  public RigidBody(double I, double I3) {
    this.I = I;
    this.I3 = I3;

    arrowX1 = new InteractiveArrow(InteractiveArrow.ARROW);
    arrowX1.setXYZ(0.0, 0.0, 0.0);
    arrowX1.setSizeXYZ(1.0, 0.0, 0.0);
    arrowX1.getStyle().setEdgeColor(Color.red);
    arrowX1.getStyle().setFillPattern(Color.red);

    traceX1 = new InteractiveTrace();
    traceX1.setConnected(true);
    traceX1.getStyle().setEdgeColor(Color.red);

    arrowX2 = new InteractiveArrow(InteractiveArrow.ARROW);
    arrowX2.setXYZ(0.0, 0.0, 0.0);
    arrowX2.setSizeXYZ(0.0, 1.0, 0.0);
    arrowX2.getStyle().setEdgeColor(Color.blue);
    arrowX2.getStyle().setFillPattern(Color.blue);

    traceX2 = new InteractiveTrace();
    traceX2.setConnected(true);
    traceX2.getStyle().setEdgeColor(Color.blue);

    arrowX3 = new InteractiveArrow(InteractiveArrow.ARROW);
    arrowX3.setXYZ(0.0, 0.0, 0.0);
    arrowX3.setSizeXYZ(0.0, 0.0, 1.0);
    arrowX3.getStyle().setEdgeColor(Color.black);
    arrowX3.getStyle().setFillPattern(Color.BLACK);

    traceX3 = new InteractiveTrace();
    traceX3.setConnected(true);
    traceX3.getStyle().setEdgeColor(Color.black);

    arrowX = new InteractiveArrow(InteractiveArrow.ARROW);
    arrowX.setXYZ(0.0, 0.0, 0.0);
    arrowX.setSizeXYZ(1.0, 0.0, 0.0);

    traceX = new InteractiveTrace();
    traceX.setConnected(true);

    arrowY = new InteractiveArrow(InteractiveArrow.ARROW);
    arrowY.setXYZ(0.0, 0.0, 0.0);
    arrowY.setSizeXYZ(0.0, 1.0, 0.0);

    traceY = new InteractiveTrace();
    traceY.setConnected(true);

    arrowZ = new InteractiveArrow(InteractiveArrow.ARROW);
    arrowZ.setXYZ(0.0, 0.0, 0.0);
    arrowZ.setSizeXYZ(0.0, 0.0, 1.0);

    traceZ = new InteractiveTrace();
    traceZ.setConnected(true);

    arrowOmega = new InteractiveArrow(InteractiveArrow.ARROW);
    arrowOmega.setXYZ(0.0, 0.0, 0.0);
    arrowOmega.setSizeXYZ(0.0, 0.0, 1.0);
    arrowOmega.getStyle().setEdgeColor(Color.orange);
    arrowOmega.getStyle().setFillPattern(Color.orange);

    traceOmega = new InteractiveTrace();
    traceOmega.setConnected(true);
    traceOmega.getStyle().setEdgeColor(Color.orange);

    arrowL = new InteractiveArrow(InteractiveArrow.ARROW);
    arrowL.setXYZ(0.0, 0.0, 0.0);
    arrowL.setSizeXYZ(0.0, 0.0, 1.0);
    arrowL.getStyle().setEdgeColor(new Color(202,67,2));
    arrowL.getStyle().setFillPattern(new Color(202,67,2));

    traceL = new InteractiveTrace();
    traceL.setConnected(true);
    traceL.getStyle().setEdgeColor(new Color(202,67,2));

    txtL = new InteractiveText("L");
    txtL.getStyle().setFont(new Font("SansSerif", Font.BOLD, 10));
    txtL.getStyle().setEdgeColor(traceL.getStyle().getEdgeColor());
    txtL.getStyle().setEdgeColor(traceL.getStyle().getEdgeColor());

    txtOmega = new InteractiveText("\u03C9");
    txtOmega.getStyle().setFont(new Font("SansSerif", Font.BOLD, 14));
    txtOmega.getStyle().setEdgeColor(traceOmega.getStyle().getEdgeColor());

    txtX1 = new InteractiveText("x1");
    txtX1.getStyle().setFont(new Font("SansSerif", Font.BOLD, 10));
    txtX1.getStyle().setEdgeColor(traceX1.getStyle().getEdgeColor());

    txtX2 = new InteractiveText("x2");
    txtX2.getStyle().setFont(new Font("SansSerif", Font.BOLD, 10));
    txtX2.getStyle().setEdgeColor(traceX2.getStyle().getEdgeColor());

    txtX3 = new InteractiveText("x3");
    txtX3.getStyle().setFont(new Font("SansSerif", Font.BOLD, 10));
    txtX3.getStyle().setEdgeColor(traceX3.getStyle().getEdgeColor());

    txtX = new InteractiveText("x");
    txtX.getStyle().setFont(new Font("SansSerif", Font.BOLD, 10));

    txtY = new InteractiveText("y");
    txtY.getStyle().setFont(new Font("SansSerif", Font.BOLD, 10));

    txtZ = new InteractiveText("z");
    txtZ.getStyle().setFont(new Font("SansSerif", Font.BOLD, 10));

    add(arrowX1);
    add(traceX1);

    add(arrowX2);
    add(traceX2);

    add(arrowX3);
    add(traceX3);

    add(arrowX);
    add(traceX);

    add(arrowY);
    add(traceY);

    add(arrowZ);
    add(traceZ);

    add(arrowOmega);
    add(traceOmega);

    add(arrowL);
    add(traceL);

    add(txtL);
    add(txtOmega);
    add(txtX1);
    add(txtX2);
    add(txtX3);
    add(txtX);
    add(txtY);
    add(txtZ);

    solver = new RK45MultiStep(this);
    solver.initialize(0.01);
    solver.setTolerance(1e-5);

  }

  /**
   * Finalizer. Does the resources clean-up.
   */
  public void finalize() {
    removeAll();
  }

  /**
   * Abstract method to be defined in subclasses. It will include also the RIGID BODY visibility.
   */
  public abstract void setVisibility();

  /**
   * Sets visibility of all objects EXCEPT rigid body according to boolean values showBodyFixedAxes, show...
   */
  public void setVisibilityExceptRigidBody() {
    arrowX1.setVisible(showBodyFixedAxes);
    arrowX2.setVisible(showBodyFixedAxes);
    arrowX3.setVisible(showBodyFixedAxes);
    txtX1.setVisible(showBodyFixedAxes);
    txtX2.setVisible(showBodyFixedAxes);
    txtX3.setVisible(showBodyFixedAxes);

    arrowX.setVisible(showLaboratoryAxes);
    arrowY.setVisible(showLaboratoryAxes);
    arrowZ.setVisible(showLaboratoryAxes);
    txtX.setVisible(showLaboratoryAxes);
    txtY.setVisible(showLaboratoryAxes);
    txtZ.setVisible(showLaboratoryAxes);

    arrowOmega.setVisible(showOmegaL);
    arrowL.setVisible(showOmegaL);
    txtOmega.setVisible(showOmegaL);
    txtL.setVisible(showOmegaL);

    if(showTraces){
      startTracePsi=getPsi();
    }

    if(showBodyFixedAxes){
      traceX1.setVisible(showTraces);
      traceX2.setVisible(showTraces);
      traceX3.setVisible(showTraces);
    }
    else {
      traceX1.setVisible(false);
      traceX2.setVisible(false);
      traceX3.setVisible(false);
    }

    if(showLaboratoryAxes){
      traceX.setVisible(showTraces);
      traceY.setVisible(showTraces);
      traceZ.setVisible(showTraces);
    }
    else{
      traceX.setVisible(false);
      traceY.setVisible(false);
      traceZ.setVisible(false);
    }

    if(showOmegaL){
      traceOmega.setVisible(showTraces);
      traceL.setVisible(showTraces);
    }
    else{
      traceOmega.setVisible(false);
      traceL.setVisible(false);
    }
  }

  /**
   * Transforms plate into the position given be three Euler's angles theta, phi and psi.
   * @param theta
   * @param phi
   * @param psi
   */
  public void transform(double theta, double phi, double psi) {
    double sinPhi = Math.sin(phi);
    double cosPhi = Math.cos(phi);
    double sinTheta = Math.sin(theta);
    double cosTheta = Math.cos(theta);
    double sinPsi = Math.sin(psi);
    double cosPsi = Math.cos(psi);

    //calculates rotation matrix

    rm[0][0] = cosPsi * cosPhi - sinPsi * cosTheta * sinPhi;
    rm[0][1] = -sinPsi * cosPhi - cosPsi * cosTheta * sinPhi;
    rm[0][2] = sinTheta * sinPhi;
    rm[1][0] = cosPsi * sinPhi + sinPsi * cosTheta * cosPhi;
    rm[1][1] = -sinPsi * sinPhi + cosPsi * cosTheta * cosPhi;
    rm[1][2] = -sinTheta * cosPhi;
    rm[2][0] = sinPsi * sinTheta;
    rm[2][1] = cosPsi * sinTheta;
    rm[2][2] = cosTheta;

    for (int i = 0; i < 3; i++) {
      for (int j = 0; j < 3; j++) {
        rmT[i][j] = rm[j][i];
      }
    }


    if (showTraces) {
      numberOfCyclesOld=numberOfCycles;
      numberOfCycles = (int)Math.floor((float)((getPsi()-startTracePsi)/(5*2.0*Math.PI)));
      if ( (((numberOfCycles-numberOfCyclesOld) == 1)||((numberOfCycles-numberOfCyclesOld) == -1)) ){
        startTracePsi=getPsi();
        clearTraces();
      }
    }

    if (!isBodyFixedFrame) {
      //components of X3, X1 and X2 vectors
      arrowX3.setSizeXYZ(rm[0][2], rm[1][2], rm[2][2]);
      if(showTraces && showBodyFixedAxes) traceX3.addPoint(rm[0][2], rm[1][2], rm[2][2]);

      arrowX1.setSizeXYZ(rm[0][0], rm[1][0], rm[2][0]);
      if(showTraces && showBodyFixedAxes) traceX1.addPoint(rm[0][0], rm[1][0], rm[2][0]);

      arrowX2.setSizeXYZ(rm[0][1], rm[1][1], rm[2][1]);
      if(showTraces && showBodyFixedAxes) traceX2.addPoint(rm[0][1], rm[1][1], rm[2][1]);

      txtX1.setXYZ(1.1 * rm[0][0], 1.1 * rm[1][0], 1.1 * rm[2][0]);
      txtX2.setXYZ(1.1 * rm[0][1], 1.1 * rm[1][1], 1.1 * rm[2][1]);
      txtX3.setXYZ(1.1 * rm[0][2], 1.1 * rm[1][2], 1.1 * rm[2][2]);

      //components of x, y and z unit vectors
      arrowZ.setSizeXYZ(0, 0, 1);

      arrowX.setSizeXYZ(1,0,0);

      arrowY.setSizeXYZ(0,1,0);

      txtX.setXYZ(1.1 * 1, 0, 0);
      txtY.setXYZ(0, 1.1 * 1, 0);
      txtZ.setXYZ(0,0, 1.1 * 1);

      //components of omega  and L vector
      double om2 = getPhiD() * cosPsi * sinTheta - getThetaD() * sinPsi;
      double om3 = getPsiD() + getPhiD() * cosTheta;
      double om1 = getPhiD() * sinPsi * sinTheta + getThetaD() * cosPsi;
      om1 = om1 / 10;
      om2 = om2 / 10;
      om3 = om3 / 10;

      double L1 = I * om1 * 10;
      double L2 = I * om2 * 10;
      double L3 = I3 * om3 * 10;

      //transformed components of omega vector
      double om1Transformed = rm[0][0] * om1 + rm[0][1] * om2 + rm[0][2] * om3;
      double om2Transformed = rm[1][0] * om1 + rm[1][1] * om2 + rm[1][2] * om3;
      double om3Transformed = rm[2][0] * om1 + rm[2][1] * om2 + rm[2][2] * om3;

      txtOmega.setXYZ(1.05 * om1Transformed, 1.05 * om2Transformed,
                      1.05 * om3Transformed);
      arrowOmega.setSizeXYZ(om1Transformed, om2Transformed, om3Transformed);
      if(showTraces && showOmegaL) traceOmega.addPoint(om1Transformed, om2Transformed, om3Transformed);

      //transformed components of L vector
      double L1Transformed = rm[0][0] * L1 + rm[0][1] * L2 + rm[0][2] * L3;
      double L2Transformed = rm[1][0] * L1 + rm[1][1] * L2 + rm[1][2] * L3;
      double L3Transformed = rm[2][0] * L1 + rm[2][1] * L2 + rm[2][2] * L3;

      arrowL.setSizeXYZ(L1Transformed, L2Transformed, L3Transformed);
      if(showTraces && showOmegaL) traceL.addPoint(L1Transformed, L2Transformed, L3Transformed);
      txtL.setXYZ(1.05 * L1Transformed, 1.05 * L2Transformed,
                  1.05 * L3Transformed);

      transformBodyOnly(rm);
    }
    else {//in body-fixed frame

      //components of X3, X1 and X2 vectors
      arrowX3.setSizeXYZ(0,0,1);

      arrowX1.setSizeXYZ(1,0,0);

      arrowX2.setSizeXYZ(0,1,0);

      txtX1.setXYZ(1.1 * 1, 0, 0);
      txtX2.setXYZ(0, 1.1 * 1, 0);
      txtX3.setXYZ(0, 0, 1.1 * 1);

      //components of x, y and z unit vectors
      arrowZ.setSizeXYZ(rmT[0][2], rmT[1][2], rmT[2][2]);
      if(showTraces && showLaboratoryAxes) traceZ.addPoint(rmT[0][2], rmT[1][2], rmT[2][2]);

      arrowX.setSizeXYZ(rmT[0][0], rmT[1][0], rmT[2][0]);
      if(showTraces && showLaboratoryAxes) traceX.addPoint(rmT[0][0], rmT[1][0], rmT[2][0]);

      arrowY.setSizeXYZ(rmT[0][1], rmT[1][1], rmT[2][1]);
      if(showTraces && showLaboratoryAxes) traceY.addPoint(rmT[0][1], rmT[1][1], rmT[2][1]);

      txtX.setXYZ(1.1 * rmT[0][0], 1.1 * rmT[1][0], 1.1 * rmT[2][0]);
      txtY.setXYZ(1.1 * rmT[0][1], 1.1 * rmT[1][1], 1.1 * rmT[2][1]);
      txtZ.setXYZ(1.1 * rmT[0][2], 1.1 * rmT[1][2], 1.1 * rmT[2][2]);


      //components of omega  and L vector
      double om2 = getPhiD() * cosPsi * sinTheta - getThetaD() * sinPsi;
      double om3 = getPsiD() + getPhiD() * cosTheta;
      double om1 = getPhiD() * sinPsi * sinTheta + getThetaD() * cosPsi;
      om1 = om1 / 10;
      om2 = om2 / 10;
      om3 = om3 / 10;

      double L1 = I * om1*10;
      double L2 = I * om2*10;
      double L3 = I3 * om3*10;

      txtOmega.setXYZ(1.05 * om1, 1.05 * om2, 1.05 * om3);
      arrowOmega.setSizeXYZ(om1, om2, om3);
      if(showTraces && showOmegaL) traceOmega.addPoint(om1, om2, om3);

      arrowL.setSizeXYZ(L1, L2, L3);
      if(showTraces && showOmegaL) traceL.addPoint(L1, L2, L3);
      txtL.setXYZ(1.05 * L1, 1.05 * L2, 1.05 * L3);

      transformBodyOnly(new double[][]{{1,0,0},{0,1,0},{0,0,1}});
    }

  }

  /**
   * Abstract class defined in subclasses. It transforms the body alone.
   */
  public abstract void transformBodyOnly(double[][] rm);

  /**
   * Makes one step
   * @return dt - delta t
   */
  public double step() {
    return solver.step();
  }

  public void doStep() {
    step();
    transform(getTheta(), getPhi(), getPsi());
  }

  /**
   * Cleares traces
   */
  public void clearTraces() {
    traceX1.clear();
    traceX2.clear();
    traceX3.clear();
    traceX.clear();
    traceY.clear();
    traceZ.clear();
    traceOmega.clear();
    traceL.clear();
  }

  public double[] getState() {
    return state;
  }

  /**
   * returns rate values
   * @param state
   * @param rate
   */
  public void getRate(double[] state, double[] rate) {
    double cosTheta, sinTheta;

    cosTheta = Math.cos(state[0]);
    sinTheta = Math.sin(state[0]);

    rate[0] = state[3];
    rate[1] = state[4];
    rate[2] = state[5];
    rate[3] = state[4] * ( (I - I3) * cosTheta * state[4] - I3 * state[5]) *
        sinTheta / I;
    rate[4] = -state[3] * ( (2 * I - I3) * cosTheta * state[4] - I3 * state[5]) /
        (sinTheta * I);
    rate[5] = state[3] *
        ( -I3 * cosTheta / sinTheta * state[5] +
         state[4] *
         ( (2 * I - I3) * cosTheta * cosTheta / sinTheta + I * sinTheta)) / I;
    rate[6] = 1;
  }

  /**
   * Sets the initial conditions for the simulation and transforms plate to its initial position.
   * Also it clears all the traces if any.
   * @param theta Euler angle in radians
   * @param phi Euler angle in radians
   * @param psi Euler angle in radians
   * @param thetaDot d(theta)/dt in rad/s
   * @param phiDot d(phi)/dt in rad/s
   * @param psiDot d(psi)/dt in rad/s
   */
  public void setInitialConditions(double theta, double phi, double psi,
                                   double thetaDot, double phiDot,
                                   double psiDot) {
    state[0] = theta;
    state[1] = phi;
    state[2] = psi;
    state[3] = thetaDot;
    state[4] = phiDot;
    state[5] = psiDot;
    state[6] = 0;
    transform(theta, phi, psi);
    clearTraces();
  }

  /**
   * Returns the Euler's theta angle.
   * @return theta Euler angle in radians
   */
  public double getTheta() {
    return state[0];
  }

  /**
   * Returns Euler angle phi
   * @return phi Euler angle in radians
   */
  public double getPhi() {
    return state[1];
  }

  /**
   * Returns the Euler's psi angle.
   * @return psi Euler angle in radians
   */
  public double getPsi() {
    return state[2];
  }

  /**
   * Returns d(theta)/dt
   * @return thetaDot in rad/s
   */
  public double getThetaD() {
    return state[3];
  }

  /**
   * Returns d(phi)/dt
   * @return phiDot in rad/s
   */
  public double getPhiD() {
    return state[4];
  }

  /**
   * Returns d(psi)/dt
   * @return psiDot in rad/s
   */
  public double getPsiD() {
    return state[5];
  }

  public double getT(){
    return state[6];
  }
}